import re
from datetime import datetime
from io import StringIO
import os
from pathlib import Path
import shlex
from typing import Union

import pandas as pd
from tqdm import tqdm


# Regex to match the find -ls output line
_RE_LINE_PATTERN = re.compile(
    r"(?P<inode>\d+)\s+"
    r"(?P<blocks>\d+)\s+"
    r"(?P<permissions>[\w-]{10})\s+"
    r"(?P<hardlinks>\d+)\s+"
    r"(?P<owner>\S+)\s+"
    r"(?P<group>\S+)\s+"
    r"(?P<size>\d+)\s+"
    r"(?P<month>\w{3})\s+"
    r"(?P<day>\d{1,2})\s+"
    r"(?P<time_or_year>\d{2}:\d{2}|\d{4})\s+"
    r"(?P<filename>.+?)"
    r"(?:\s+->\s+(?P<symlink_target>.+))?$"
)


def _parse_line(line: str):
    # pylint: disable=W0718

    match = _RE_LINE_PATTERN.match(line.strip())
    if not match:
        return None
    parts = match.groupdict()

    # Parse datetime
    month = parts["month"]
    day = parts["day"]
    time_or_year = parts["time_or_year"]
    try:
        if ":" in time_or_year:
            # No year in output, use current year
            dt_str = f"{month} {day} {datetime.now().year} {time_or_year}"
            dt = datetime.strptime(dt_str, "%b %d %Y %H:%M")
        else:
            dt_str = f"{month} {day} {time_or_year}"
            dt = datetime.strptime(dt_str, "%b %d %Y")
    except Exception:
        dt = pd.NaT

    # Attempt shell split to get the full path
    try:
        full_path = shlex.split(parts["filename"].strip())[0]
    except ValueError:
        full_path = parts["filename"].strip()

    # Calculate directory and filename
    directory = os.path.dirname(full_path)
    filename = os.path.basename(full_path)

    return {
        "Inode": int(parts["inode"]),
        "Blocks": int(parts["blocks"]),
        "Permissions": parts["permissions"],
        "Hardlinks": int(parts["hardlinks"]),
        "Owner": parts["owner"],
        "Group": parts["group"],
        "Filesize": int(parts["size"]),
        "Datetime": dt,
        "Parent": directory,
        "Filename": filename,
        "SymlinkTarget": parts["symlink_target"],
        "IsSymlink": parts["permissions"].startswith("l"),
    }


def _parse_iterable(iterable):
    return pd.DataFrame(
        [parsed for line in iterable if (parsed := _parse_line(line))]
    )


def read_filetree(file: Union[str, Path, StringIO], common: bool = True) -> pd.DataFrame:
    """
    Reads the output of `find -ls` command and returns a DataFrame.
    Handles symlinks and combines Month, Day, Time/Year into a single Datetime column.
    """
    if isinstance(file, (str, Path)):
        with open(file, "r", encoding="utf-8") as f:
            df = _parse_iterable(f).sort_values(
                "Datetime", ignore_index=True
            )

    elif isinstance(file, StringIO):
        df = _parse_iterable(file).sort_values(
            "Datetime", ignore_index=True
        )
    else:
        raise TypeError(f"'file' must be str, Path, or StringIO, not {type(file).__name__}")

    if common:
        value = os.path.commonprefix(list(df["Filename"]))
        df["Filename"] = df["Filename"].apply(
            lambda x: x.removeprefix(value).removeprefix("/")
        )
    return df


def compute_directory_size(df, only_directories=False):
    """
    Compute directory sizes and their hierarchical relationships.

    Parameters:
        df (pd.DataFrame): DataFrame with 'Parent' and 'Filename' columns
        only_directories (bool): If True, only show directories in the visualization

    Returns:
        tuple: (sizes, parents) where:
            - sizes: dict mapping paths to their total sizes
            - parents: dict mapping paths to their parent directories
    """
    if df.empty:
        return {}, {}

    total_sizes = {}
    parents = {}

    # Step 1: Process all files, recording their size and parent.
    for _, row in tqdm(df.iterrows(), total=len(df), leave=False, desc="Processing files"):
        parent_path = row["Parent"]
        file_path = os.path.join(parent_path, row["Filename"])
        total_sizes[file_path] = int(row["Filesize"])
        parents[file_path] = parent_path

    # Step 2: Ensure all directories and their ancestors are in the data structures.
    all_paths = set(parents.keys())
    for path in all_paths:
        parent = parents.get(path)
        while parent and parent not in parents:
            if parent not in total_sizes:
                total_sizes[parent] = 0

            grandparent = os.path.dirname(parent)
            if grandparent == parent:  # Reached the root
                parents[parent] = ""
                break
            parents[parent] = grandparent
            parent = grandparent

    # Step 3: Propagate sizes up the tree from the longest paths to the shortest.
    for path in sorted(total_sizes.keys(), key=len, reverse=True):
        parent = parents.get(path)
        if parent is not None and parent in total_sizes:
            total_sizes[parent] += total_sizes[path]

    # If only directories are requested, filter out the files.
    if only_directories:
        # A directory is any path that is a parent of another path.
        dir_paths = set(parents.values())

        # Filter total_sizes and parents to only include directories.
        total_sizes = {
            path: size for path, size in total_sizes.items() if path in dir_paths
        }
        parents = {
            path: parent for path, parent in parents.items() if path in dir_paths
        }

    # Any node whose parent is not in the dataset is a root.
    all_paths = set(total_sizes.keys())
    for path, parent in parents.items():
        if parent not in all_paths:
            parents[path] = ""

    return total_sizes, parents
